/*

2014-2020 by Andreas Romeyke (SLUB Dresden)

The code contributed by SLUB is licensed under apache 2.0 and based partially
on NFS Storage Plugin, 

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package org.slub.rosetta.dps.repository.plugin.storage.nfs;

import com.exlibris.core.infra.common.exceptions.logging.ExLogger;
import com.exlibris.core.infra.common.util.IOUtil;
import com.exlibris.core.infra.svc.api.scriptRunner.ExecExternalProcess;
import com.exlibris.core.sdk.storage.containers.StoredEntityMetaData;
import com.exlibris.core.sdk.storage.handler.AbstractStorageHandler;
import com.exlibris.core.sdk.storage.handler.StorageUtil;
import com.exlibris.core.sdk.utils.FileTransferUtil;
import com.exlibris.core.sdk.utils.FileUtil;
import com.exlibris.digitool.common.dnx.DnxDocument;
import com.exlibris.digitool.common.dnx.DnxSection;
import com.exlibris.digitool.common.dnx.DnxSectionRecord;
import com.exlibris.digitool.common.storage.Fixity;
import com.exlibris.digitool.infrastructure.utils.Checksummer;

import java.io.BufferedInputStream;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.nio.channels.FileChannel;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.Paths;
import java.security.NoSuchAlgorithmException;
import java.text.SimpleDateFormat;
import java.util.ArrayList;
import java.util.Calendar;
import java.util.Date;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

/**
 * SLUBStoragePlugin
 * writes all IEs, files etc. into the same directory under yyyy/mm/dd/IEpid/
 *
 * HINT: this plugin is *only* useful to store AIPs (IEs + files) in
 *       permanent storage. Otherwise (e.g in the case of SIPs) it can't
 *       detect the IE-PID and will report an error.
 *
 * the fixities will only be checked if files are stored (storeEntity()), 
 * because the Rosetta API does not give enough information about available
 * fixities by calling retrieveEntity() to read files, because only 
 * a string containing the path is given. 
 *
 * Please note, that all undocumented codelines are in most cases copied from
 * com.exlibris.dps.repository.plugin.storage.nfs.NFSStoragePlugin and I am
 * unable to find full documentation.
 *
 * @author andreas.romeyke@slub-dresden.de (Andreas Romeyke)
 * @ see com.exlibris.dps.repository.plugin.storage.nfs.NFSStoragePlugin
 */
public class SLUBStoragePlugin extends AbstractStorageHandler {
    private static final ExLogger log = ExLogger.getExLogger(SLUBStoragePlugin.class);

    private String getDirRoot () {
        String path = parameters.get("DIR_ROOT");
        if (null == path) {
            log.error("Could not retrieve DIR_ROOT");
        }
        return path;
    }
    private int getBlockSize () {
        int blockSize = 32*1024;
        String blocksizeStr = this.parameters.get("BLOCK_SIZE");
        if (null != blocksizeStr) {
            try {
                blockSize = Integer.parseInt(blocksizeStr);
            } catch (NumberFormatException e) {
                log.error("Could not convert BLOCK_SIZE string to int, " + e.getMessage());
            }
        } else {
            log.error("Could not retrieve BLOCK_SIZE");
        }
        return blockSize;
    }
    private String getFilesHandlingMethod () {
        String filehandlingMethod = this.parameters.get("FILES_HANDLING_METHOD");
        if (null == filehandlingMethod) {
            log.error("Could not retrieve FILES_HANDLING_METHOD");
        }
        return filehandlingMethod;
    }

    private String throughput(long starttime_in_ms, long endtime_in_ms, String filename) {
        try {
            long fsize = Files.size(Paths.get(filename));
            return throughput(starttime_in_ms, endtime_in_ms, fsize);
        } catch (IOException e) {
            log.warn("SLUBStoragePlugin.throughput, " + e.getMessage());
            /* do nothing */
        }
        return "(unknown) Bytes/s";
    }
    private String throughput(long starttime_in_ms, long endtime_in_ms, long fsize) {
        long duration_in_ms = endtime_in_ms - starttime_in_ms;
        //log.debug("SLUBStoragePlugin.throughput duration=" + duration_in_ms + " filesize=" + fsize);
        if ((fsize >= 0) && (duration_in_ms) > 0) {
            /* fsize in Bytes
               duration in ms = duration in s / 1000
               ->
               fsize_B/duration_in_s = fsize_B*1000/duration_in_ms
               ->
               fsize_MB/duration_in_s = fsize_B*1000/(1024*1024*duration_in_s)

               Example:
               fsize_B=31561376
               duration_in_ms = 86
               ->
               386MB/s
             */
            return (fsize * 1000) / (1024 * 1024 * duration_in_ms) + " MBytes/s";
        }
        return "(unknown) Bytes/s";
    }

    /**
     * checkFixity
     * @param fixities
     * @param storedEntityIdentifier (relative filepath)
     * @return true if valid, otherwise false
     * @throws Exception
     */
    public boolean checkFixity(List<Fixity> fixities, String storedEntityIdentifier) throws Exception {
        //log.debug("SLUBStoragePlugin.checkFixity() storedEntityIdentifier='" + storedEntityIdentifier + "'");
        String absolute_storedEntityIdentifier = getFullFilePath(storedEntityIdentifier);
        contractAssertIsAbsolutePath(absolute_storedEntityIdentifier);
        //log.debug("SLUBStoragePlugin.checkFixity() absolute_storedEntityIdentifier='" + absolute_storedEntityIdentifier + "'");
        if (null==fixities) {
            log.warn("SLUBStoragePlugin.checkFixity() No fixity list provided");
            return true;
        }
        boolean result = true;
        for (Fixity fixity : fixities) {
            fixity.setResult(Boolean.FALSE);
            result &= checkSpecificFixity(absolute_storedEntityIdentifier, fixity);
        }
        return result;
    }

    /**
     * checkSpecificFixity
     * @param absolute_storedEntityIdentifier absolute file path
     * @param fixity
     * @return true if valid otherwise false
     * @throws Exception
     */
    private boolean checkSpecificFixity(String absolute_storedEntityIdentifier, Fixity fixity) throws Exception {
        contractAssertIsAbsolutePath(absolute_storedEntityIdentifier);
        String algorithm = fixity.getAlgorithm();
                    /* HINT: in past versions of Rosetta was a workaround for fixity.getAlgorithm() required,
                             a lowercase string was returned instead of the correct fixity code table entry */
        //log.debug("SLUBStoragePlugin.checkSpecificFixity() absolute_storedEntityIdentifier='"+absolute_storedEntityIdentifier+"' getAlgorithm=" + algorithm);
        try {
            if (
                    (!Fixity.FixityAlgorithm.MD5.toString().equals(algorithm)) &&
                            (!Fixity.FixityAlgorithm.SHA1.toString().equals(algorithm)) &&
                            (!Fixity.FixityAlgorithm.SHA256.toString().equals(algorithm)) &&
                            (!Fixity.FixityAlgorithm.CRC32.toString().equals(algorithm))
            ) {
                //log.debug("SLUBStoragePlugin.checkSpecificFixity() call checkFixityByPlugin (" + fixity + "," + absolute_storedEntityIdentifier + ")");
                return checkFixityByPlugin(fixity, absolute_storedEntityIdentifier);
            } else {
                //log.debug("SLUBStoragePlugin.checkSpecificFixity() calcMD5|calcSHA1|calcCRC32|calcSHA256=true");
                int checksummerAlgorithmIndex = getChecksummerAlgorithmIndex(algorithm);
                //log.debug("SLUBStoragePlugin.checkSpecificFixity() checksummerAlgorithmIndex=" + checksummerAlgorithmIndex);
                if (checksummerAlgorithmIndex != -1) {
                    return checkFixityByBuiltin(fixity, absolute_storedEntityIdentifier, algorithm);
                }
            }
        } catch (IOException e) {
            log.error("SLUBStoragePlugin.checkSpecificFixity(), I/O error, " + e.getMessage());
            throw e; // let Rosetta know something broke, creates technical issue in workbench
        } catch (NoSuchAlgorithmException e) {
            log.error("SLUBStoragePlugin.checkSpecificFixity(), missed Algorithm, " + e.getMessage());
            throw e; // let Rosetta know something broke, creates technical issue in workbench
        } catch (Exception e) {
            log.error("SLUBStoragePlugin.checkSpecificFixity(), unknown problem, " + e.getMessage());
            throw e; // let Rosetta know something broke, creates technical issue in workbench
        }
        return false;
    }

    /** uses a checksummer object and check if fixity of given builtin algorithm is correct
     *
     * @param fixity concrete fixity object
     * @param absolute_storedEntityIdentifier absolute filepath
     * @param algorithm which algorithm should be used
     * @return true if valid, otherwise false
     * @throws NoSuchAlgorithmException if algorithm is unknown
     * @throws IOException if I/O error
     */
    private boolean checkFixityByBuiltin(Fixity fixity, String absolute_storedEntityIdentifier, String algorithm) throws NoSuchAlgorithmException, IOException {
        //log.debug("SLUBStoragePlugin.checkFixityByBuiltin() absolute_storedEntityIdentifier='"+absolute_storedEntityIdentifier+"' algorithm=" + algorithm);
        contractAssertIsAbsolutePath(absolute_storedEntityIdentifier);
        String oldValue = fixity.getValue();
        //log.debug("SLUBStoragePlugin.checkFixityByBuiltin() getAlgorithm (2)=" + algorithm);
        //log.debug("SLUBStoragePlugin.checkFixityByBuiltin() oldvalue=" + oldValue);
        long starttime = System.currentTimeMillis();
        String storedEntityIdentifier = relativePath(absolute_storedEntityIdentifier);
        //log.debug("SLUBStoragePlugin.checkFixityByBuiltin() calculated storedEntityIdentifier=" + storedEntityIdentifier);
        contractAssertIsRelativePath(storedEntityIdentifier);
        InputStream is = retrieveEntity(storedEntityIdentifier);
        Checksummer checksummer = new Checksummer(is, true, true, true, true);
        fixity.setValue(checksummer.getChecksum(algorithm));
        var newValue = fixity.getValue();
        //log.debug("SLUBStoragePlugin.checkFixityByBuiltin() newvalue=" + newValue);
        fixity.setResult((oldValue == null) || (oldValue.equalsIgnoreCase(newValue)));
        boolean result = fixity.getResult();
        is.close();
        long endtime = System.currentTimeMillis();
        log.info("SLUBStoragePlugin.checkFixityByBuiltin() pathname='" + absolute_storedEntityIdentifier + "'"
                + " algorithm=" + algorithm
                + " oldvalue=" + oldValue
                + " newvalue=" + newValue
                + " result=" + result
                + " (" + throughput(starttime, endtime, absolute_storedEntityIdentifier)+ ")"
        );
        return result;
    }

    String relativePath(String absolute_storedEntityIdentifier) {
        var begin = getDirRoot().length();
        var end = absolute_storedEntityIdentifier.length();
        var rel = absolute_storedEntityIdentifier.substring(begin, end);
        //log.debug("SLUBStoragePlugin.relativePath('"+ absolute_storedEntityIdentifier + "') -> DirRoot="+getDirRoot() + " begin=" + begin + " end=" + end + " rel='" + rel + "'");
        assert(begin < end);
        return rel;
    }


    /** check fixity by calling its registered plugin
     *
     * @param fixity concrete fixity object
     * @param absolute_storedEntityIdentifier absolute filepath to file which should be checked
     * @return true if valid, otherwise false
     * @throws Exception to inform rosetta
     */
    private boolean checkFixityByPlugin(Fixity fixity, String absolute_storedEntityIdentifier) throws Exception {
        //log.debug("SLUBStoragePlugin.checkFixityByPlugin() another fixity, root=" + getDirRoot() + " absolute_storedEntityIdentifier=" + absolute_storedEntityIdentifier);
        contractAssertIsAbsolutePath(absolute_storedEntityIdentifier);
        String pluginname = fixity.getPluginName();
        String algorithm = "should not occur";
        //log.debug("SLUBStoragePlugin.checkFixityByPlugin() pluginname=" + pluginname);
        // HINT: Workaround for Rosetta case 06472860 - Storage Plugin: Empty custom fixity plugin name breaks file updates [Rosetta 7.3.0.0]
        //       -> use fallback names in case fixity plugin name is missing
        if (pluginname == null || pluginname.length() == 0) {
            log.warn("SLUBStoragePlugin.checkFixityByPlugin() failed to get pluginname, because it is empty. Possibly, there is no valid fixity type used or a CustomFixityPlugin missed");
            log.warn("SLUBStoragePlugin.checkFixityByPlugin() trying to use fallback table to determine plugin name");
            algorithm = fixity.getAlgorithm();
            if ("SHA512".equals(algorithm)) {
                pluginname = "CustomFixitySHA512Plugin";
                //log.debug("SLUBStoragePlugin.checkFixityByPlugin() pluginname=" + pluginname);
            } else {
                log.error("SLUBStoragePlugin.checkFixityByPlugin() fallback table has no plugin name entry for algorithm '" + algorithm + "'");
            }
        }
        String oldValue = fixity.getValue();
        log.info("SLUBStoragePlugin.checkFixityByPlugin() oldvalue=" + oldValue);
        long starttime = System.currentTimeMillis();
        String newValue = "should not occur";
        try {
            newValue = getChecksumUsingPlugin(absolute_storedEntityIdentifier, pluginname, oldValue);
        } catch  (Exception e) {
            log.error("SLUBStoragePlugin.checkFixityByPlugin() exception in getChecksumUsingPlugin("+absolute_storedEntityIdentifier + ", "+ pluginname + ", " + oldValue +"), " + e.getMessage() );
        }
        fixity.setValue(newValue);
        /* HINT: if plugin name is still empty a java.lang.NullPointerException gets thrown
                 Rosetta will handle it and let us know (creates technical issue in workbench)
         */
        fixity.setResult((oldValue == null) || (oldValue.equals(newValue)));
        boolean result = fixity.getResult();
        long endtime = System.currentTimeMillis();
        log.info("SLUBStoragePlugin.checkFixityByPlugin() pathname='" + absolute_storedEntityIdentifier + "'"
                + " algorithm=" + algorithm
                + " oldvalue=" + oldValue
                + " newvalue=" + newValue
                + " result=" + result
                + " (" + throughput(starttime, endtime, absolute_storedEntityIdentifier)+ ")"
        );
        return result;
    }

    /**
     * deleteEntity
     * @param relative_storedEntityIdentifier
     * @return true if successful
     */
    public boolean deleteEntity(String relative_storedEntityIdentifier)
    {
        log.info("SLUBStoragePlugin.deleteEntity() storedEntityIdentifier='" + relative_storedEntityIdentifier + "'");
        var absolute_filename = getFullFilePath(relative_storedEntityIdentifier);
        File file = new File(absolute_filename);
        try
        {
            return file.delete();
        }
        catch (Exception e)
        {
            log.warn("SLUBStoragePlugin.deleteEntity() failed to delete entity with path: " + file.getPath(), e.getMessage());
        }
        return true;
    }

    public String getLocalFilePath(String storedEntityIdentifier)
    {
        //log.debug("SLUBStoragePlugin.getLocalFilePath() with '" + storedEntityIdentifier + "'");
        return getFullFilePath(storedEntityIdentifier);
    }

    /**
     * getFullFilePath
     * @param storedEntityIdentifier - relative filepath
     * @return absolute filepath
     */
    public String getFullFilePath(String storedEntityIdentifier)
    {
        //log.debug("SLUBStoragePlugin.getFullFilePath() with '" + storedEntityIdentifier + "'");
        return getDirRoot() + storedEntityIdentifier;
    }

    private void contractAssertIsRelativePath(String storedEntityIdentifier) throws IOException {
        if ((storedEntityIdentifier.startsWith(getDirRoot()))) {
            String message = "SLUBStoragePlugin.contractAssertIsRelativePath() storedEntityIdentifier='" + storedEntityIdentifier + "' should be relative, but starts with ROOT '"+getDirRoot()+"'";
            log.error(message);
            throw new IOException(message);
        }
    }
    private void contractAssertIsAbsolutePath(String absolute_storedEntityIdentifier) throws IOException {
        if ((!absolute_storedEntityIdentifier.startsWith(getDirRoot()))) {
            String message = "SLUBStoragePlugin.contractAssertIsAbsolutePath() absolute_storedEntityIdentifier='" + absolute_storedEntityIdentifier + "' starts not with ROOT '" + getDirRoot() + "'";
            log.error(message);
            throw new IOException(message);
        }
        if ((absolute_storedEntityIdentifier.startsWith(getDirRoot() + getDirRoot()))) {
            String message = "SLUBStoragePlugin.contractAssertIsAbsolutePath() absolute_storedEntityIdentifier='"+ absolute_storedEntityIdentifier + "' starts with ROOT '" + getDirRoot() + "' twice!";
            log.error(message);
            throw new IOException(message);
        }
    }
    /**
     * retrieveEntity
     * @param storedEntityIdentifier as relative filepath
     * @return Buffered InputStream
     * @throws IOException
     */
    public InputStream retrieveEntity(String storedEntityIdentifier)
            throws IOException
    {
        log.info("SLUBStoragePlugin.retrieveEntity() with storedEntityIdentifier '" + storedEntityIdentifier + "'");
        contractAssertIsRelativePath(storedEntityIdentifier);
        var absolute_filename = getFullFilePath(storedEntityIdentifier);
        contractAssertIsAbsolutePath(absolute_filename);
        var absolute_path = Paths.get(absolute_filename);
        var is = java.nio.file.Files.newInputStream( absolute_path );
        return new BufferedInputStream( is , getBlockSize());
    }

    /**
     * retrieveEntityByRange
     * @param storedEntityIdentifier as relative filepath
     * @param start offset to begin
     * @param end   offset to end
     * @return byte array of size end-begin
     * @throws Exception
     */
    public byte[] retrieveEntityByRange(String storedEntityIdentifier, long start, long end) throws Exception
    {
        log.info("SLUBStoragePlugin.retrieveEntitybyRange() with storedEntityIdentifier='" + storedEntityIdentifier + "' start=" + start + " end=" + end);
        /* try-with-ressource block, no close if failed needed */
        var absolute_filename = getFullFilePath(storedEntityIdentifier);
        try (RandomAccessFile file = new RandomAccessFile(absolute_filename, "r")) {
            long starttime = System.currentTimeMillis();
            file.seek(start);
            byte[] bytes = new byte[(int)(end - start + 1L)];
            file.readFully(bytes, 0, (int)(end - start + 1L));
            long endtime = System.currentTimeMillis();
            log.info("SLUBStoragePlugin.retrieveEntitybyRange() sucessfull (" + throughput(starttime, endtime, file.length())+ ")");
            return bytes;
        } catch (FileNotFoundException e) {
            log.error("SLUBStoragePlugin.retrieveEntitybyRange(), file not found, " + e.getMessage());
            throw e;
        } catch (IOException e) {
            log.error("SLUBStoragePlugin.retrieveEntityByRange(), I/O error, " + e.getMessage());
            throw e;
        } catch (Exception e) {
            log.error("SLUBStoragePlugin.retrieveEntityByRange(), unknown problem, " + e.getMessage());
            throw e;
        }
    }

    /**
     * storeEntity
     * @param is - InputStream
     * @param storedEntityMetadata
     * @return relative path to storedEntity
     * @throws Exception
     */
    public String storeEntity(InputStream is, StoredEntityMetaData storedEntityMetadata)
            throws Exception
    {
        log.info("SLUBStoragePlugin.storeEntity()");
        String existsDescPath = getFilePathInDescIfExists(storedEntityMetadata);
        //log.debug("SLUBStoragePlugin.storeEntity() existsDescPath='" + existsDescPath + "'");
        String destFilePath = null;

        boolean isCopyFileNeeded = true;
        if (existsDescPath != null)
        {
            destFilePath = existsDescPath;
            /* destFilePath is relative, checkfixity expects absolute */
            //log.debug("SLUBStoragePlugin.storeEntity() calling checkFixity(storedEntityMetadata.getFixities(), '"+destFilePath+");");
            isCopyFileNeeded = !checkFixity(storedEntityMetadata.getFixities(), destFilePath);
        }
        //log.debug("SLUBStoragePlugin.storeEntity() destFilePath='" + destFilePath + "'");
        Map<String, String> paths = getStoreEntityIdentifier(storedEntityMetadata, destFilePath);
        String storedEntityIdentifier = paths.get("relativeDirectoryPath");
        contractAssertIsRelativePath(storedEntityIdentifier);
        //log.debug("SLUBStoragePlugin.storeEntity() storedEntityIdentifier='" + storedEntityIdentifier + "'");
        String absolute_destFilePath = paths.get("destFilePath");
        contractAssertIsAbsolutePath(absolute_destFilePath);
        //log.debug("SLUBStoragePlugin.storeEntity() destFilePath (2)='" + absolute_destFilePath + "'");
        //log.debug("SLUBStoragePlugin.storeEntity() isCopyFileNeeded='" + isCopyFileNeeded + "'");
        if (isCopyFileNeeded)
        {
            if (canHandleSourcePath(storedEntityMetadata.getCurrentFilePath()))
            {
                //log.debug("SLUBStoragePlugin.storeEntity() absolute_destFilePath canhandle sourcepath");
                if (is != null) {
                    is.close();
                }
                copyStream(storedEntityMetadata, absolute_destFilePath);
                //log.debug("SLUBStoragePlugin.storeEntity() try copy (copyStream(storedEntityMetadata, '" + absolute_destFilePath + "')) was successfull");
            }
            else
            {
                //log.debug("SLUBStoragePlugin.storeEntity() Cannot handle source path: " + storedEntityMetadata.getCurrentFilePath());
                if (is == null)
                {
                    log.warn("SLUBStoragePlugin.storeEntity() InputStream is null");
                    return null;
                }
                try (OutputStream output = java.nio.file.Files.newOutputStream( Paths.get( absolute_destFilePath)))
                {
                    long starttime = System.currentTimeMillis();
                    IOUtil.copy(is, output, getBlockSize());
                    long endtime = System.currentTimeMillis();
                    log.info("SLUBStoragePlugin.storeEntity() try copy (IOUtil.copy(is, '"+absolute_destFilePath+"', blocksize) was successfull (" + throughput(starttime, endtime, absolute_destFilePath)+ ")" );
                }
            }
            if (!checkFixity(storedEntityMetadata.getFixities(), storedEntityIdentifier)) {
                log.warn("SLUBStoragePlugin.storeEntity() called checkFixity(fixities, '"+storedEntityIdentifier+"') failed");
                return null;
            }
        }
        //log.debug("SLUBStoragePlugin.storeEntity() storedEntityIdentifier (2)='" + storedEntityIdentifier + "'");
        return storedEntityIdentifier;
    }

    /**
     * copyFile, specialised function derived from FileUtil.copyFile using improved filecopy
     * @param source
     * @param destination
     * @throws IOException
     */
    private void copyFile(String source, String destination) throws IOException {
        long starttime = System.currentTimeMillis();
        if (!source.toLowerCase().trim().startsWith("http")) { /* simpleFilecopy, see https://www.digitalocean.com/community/tutorials/java-copy-file */
            try (
                    FileChannel sourceChannel = new FileInputStream(source).getChannel();
                    FileChannel destChannel = new FileOutputStream(destination).getChannel()
            ) {
                destChannel.transferFrom(sourceChannel, 0, sourceChannel.size());
            }
        } else {
            InputStream input = FileTransferUtil.getUrlContent(source);
            FileOutputStream output = new FileOutputStream(destination);
            IOUtil.copy(input, output, getBlockSize());
            IOUtil.shutdownStream(input);
            IOUtil.shutdownStream(output);
        }
        long endtime = System.currentTimeMillis();
        log.info("SLUBStoragePlugin.copyFile (" + source + ", " + destination + ") was successfull (" + throughput(starttime, endtime, destination) + ")");
    }
    protected void copyStream(StoredEntityMetaData storedEntityMetadata, String destPath)
            throws IOException
    {
        log.info("SLUBStoragePlugin.copyStream()");
        String filesHandlingMethod = getFilesHandlingMethod();
        String srcPath = storedEntityMetadata.getCurrentFilePath();
        //log.debug("SLUBStoragePlugin.copyStream() destPath='" + destPath + "'");
        //log.debug("SLUBStoragePlugin.copyStream() srcPath='" + srcPath + "'");
        //log.debug("SLUBStoragePlugin.copyStream() filesHandlingMethod='" + filesHandlingMethod + "'");
        String pid = storedEntityMetadata.getEntityPid();
        //log.debug("SLUBStoragePlugin.copyStream() pid='" + pid + "'");
        String iePid = storedEntityMetadata.getIePid();
        //log.debug("SLUBStoragePlugin.copyStream() iePid='" + iePid + "'");
        String attr = "('\"+srcPath+\"','\" + destPath + \"')";
        if ("copy".equalsIgnoreCase(filesHandlingMethod)) {
            // FileUtil.copyFile(srcPath, destPath);
            copyFile(srcPath, destPath);
            saveDestPathsTmpFile(iePid, pid, destPath);
            log.info("SLUBStoragePlugin.copyStream(), copy"+attr + " was successful");
        } else if ("move".equalsIgnoreCase(filesHandlingMethod)) {
            File canonicalSrcFile = getCanonicalFile(srcPath);
            assert canonicalSrcFile != null;
            FileUtil.moveFile(canonicalSrcFile, new File(destPath));
            saveDestPathsTmpFile(iePid, pid, destPath);
            log.info("SLUBStoragePlugin.copyStream(), move"+attr+" was successful");
        }
        else if ("soft_link".equalsIgnoreCase(filesHandlingMethod)) {
            softLink(srcPath, destPath);
            log.info("SLUBStoragePlugin.copyStream(), softlink"+attr+" was successful");
        }
        else if ("hard_link".equalsIgnoreCase(filesHandlingMethod)) {
            hardLink(srcPath, destPath);
            log.info("SLUBStoragePlugin.copyStream(), hardlink"+attr+" was successful");
        }
        else {
            String msg = "SLUBStoragePlugin.copyStream(), unknown filehandling method detected: " + filesHandlingMethod;
            log.fatal(msg);
            throw new IOException(msg);
        }
    }
    protected String getFilePathInDescIfExists(StoredEntityMetaData storedEntityMetadata)
    {
        log.info("SLUBStoragePlugin.getFilePathInDescIfExists()");
        if (storedEntityMetadata.getIePid() == null) {
            return null;
        }
        String tmpFilePath = getTempStorageDirectory(false) + "destPath";
        String existsDescPath = StorageUtil.readDestPathFromTmpFile(storedEntityMetadata.getIePid(), tmpFilePath, storedEntityMetadata.getEntityPid());
        //log.debug("SLUBStoragePlugin.getFilePathInDescIfExists() existsDescPath='" + existsDescPath + "'");
        return existsDescPath;
    }

    /** copied from NFS Storage Plugin, enhanced with debugging info, {@inheritDoc} */
    private boolean canHandleSourcePath(String srcPath) {
        //log.debug("SLUBStoragePlugin.canHandleSourcePath path='" + srcPath + "'");
        try {
            File file = new File(srcPath);
            return file.canRead();
        }
        catch (Exception e) {
            // log.debug("SLUBStoragePlugin.canHandleSourcePath path is empty (path='" + srcPath + "' exception=" + e + ")");
            return false;
        }
    }

    /** copied from NFS Storage Plugin, enhanced with debugging info, {@inheritDoc} */
    private int getChecksummerAlgorithmIndex(String algorithm)
    {
        //log.debug("SLUBStoragePlugin.getChecksummerAlgorithm() algorithm='" + algorithm + "'");
        try
        {
            Fixity.FixityAlgorithm fixityAlgorithm = Fixity.FixityAlgorithm.valueOf(algorithm);
            return fixityAlgorithm.ordinal();
        }
        catch (Exception e) {
            log.info("SLUBStoragePlugin.getChecksummerAlgorithm(), exception: " + e.getMessage());
        }
        return -1;
    }




    /** copied from NFS Storage Plugin, enhanced with debugging info, {@inheritDoc} */
    private File getCanonicalFile(String srcPath)
    {
        //log.debug("SLUBStoragePlugin.getCanonicalFile() srcPath='"+ srcPath + "'");
        String backslash = "\\";
        String fileName = srcPath.split(backslash + File.separator)[(srcPath.split(backslash + File.separator).length - 1)];
        try
        {
            File canonicalSrcDir = new File(srcPath).getParentFile().getCanonicalFile();
            return new File(canonicalSrcDir, fileName).getCanonicalFile();
        }
        catch (IOException e) {
            log.warn("getCanonicalFile of '" + srcPath + "':" + e.getMessage());
        }
        return null;
    }

    /** copied from NFS Storage Plugin, enhanced with debugging info, {@inheritDoc} */
    private Map<String, String> getStoreEntityIdentifier(StoredEntityMetaData storedEntityMetadata, String destFilePath) throws Exception
    {
        log.info("SLUBStoragePlugin.getStoreEntityIdentifier() destFilePath='" + destFilePath +"'");
        Map<String, String> paths = new HashMap<>();
        //log.debug("(1) storedEntityMetadata is null?" + (null == storedEntityMetadata));
        String fileName = createFileName(storedEntityMetadata);
        //log.debug("fileName='" + fileName + "'");
        //log.debug("(2) storedEntityMetadata is null?" + (null == storedEntityMetadata));
        String relativeDirectoryPath = getStreamRelativePath(storedEntityMetadata);
        //log.debug("relativeDirectoryPath='" + relativeDirectoryPath + "'");
        if (destFilePath == null)
        {
            File destFile = getStreamDirectory(relativeDirectoryPath, fileName);
            destFilePath = destFile.getAbsolutePath();

        }
        contractAssertIsAbsolutePath(destFilePath);
        //log.debug("destFilePath (2)='" + destFilePath + "'");
        paths.put("destFilePath", destFilePath);
        String relativePath = relativeDirectoryPath + File.separator + fileName;
        contractAssertIsRelativePath(relativePath);
        paths.put("relativeDirectoryPath", relativePath);
        log.info(
                "SLUBStoragePlugin.getStoreEntityIdentifier() stored: "
                + "destFilePath='" + destFilePath + "'"
                + "relativeDirectoryPath='" + relativePath + "'"
        );
        return paths;
    }
    /** copied from NFS Storage Plugin, enhanced with debugging info,
     * this combines full file path and creates parent directories if needed  {@inheritDoc}
     */
    File getStreamDirectory(String path, String fileName) {
        log.info("SLUBStoragePlugin.getStreamDirectory path='" + path + "' fileName='" + fileName + "'");
        File newDir = new File(getDirRoot() + File.separator + path);
        //log.debug("SLUBStoragePlugin.getStreamDirectory newDir.getAbsolutePath()=" + newDir.getAbsolutePath());
        boolean arecreated = newDir.mkdirs();
        //log.debug("SLUBStoragePlugin.getStreamDirectory newDir.mkdirs(), directories are created:" + arecreated);
        return new File(newDir.getAbsolutePath() + File.separator + fileName);
    }
    /** prepare right path
     * path should be of form yyyy/MM/dd/IE-PID/
     * we need to findout the associated dnx document (IE),
     * get the creation date of the IE/SIP
     * and get the IE PID of the associated IE.
     * @param storedEntityMetaData StoredEntityMetaData
     * @return calculated relative path as String
     * returns the path as string
     */

    String getStreamRelativePath(StoredEntityMetaData storedEntityMetaData) throws Exception {
        log.info("SLUBStoragePlugin.getStreamRelativePath()");
        if (null == storedEntityMetaData) {
            log.fatal("SLUBStoragePlugin.getStreamRelativePath() with empty storedEntityMetaData called!");
            throw new AssertionError();
        }
        //log.debug("SLUBStoragePlugin.getStreamRelativePath() assert fine");
        String relativeDirectoryPath = File.separator;
        // get IE PID by calling IE-DNX record and search for ""internalIdentifierType" == "PID"
        DnxDocument iedoc = storedEntityMetaData.getIeDnx();
        //log.debug("SLUBStoragePlugin.getStreamRelativePath() getIeDnx fine");
        if (null == iedoc) {
            log.error ("SLUBStoragePlugin.getStreamRelativePath no iedoc found, do you use plugin for others than permanent data (IE and filestream)? You should not! Also do not use it for SIPstoragePermanentGroup!");
            throw new Exception("error, no iedoc found, do you use plugin for others than permanent data (IE and filestream)? You should not! Also do not use it for SIPstoragePermanentGroup!");
        }
        StoredEntityMetaData.EntityType entityType = storedEntityMetaData.getEntityType();
        //log.debug("SLUBStoragePlugin.getStreamRelativePath() getEntityType fine");
        String entitytype = entityType.name();
        //log.debug("entitytype='" + entitytype + "'");
        DnxSection iesec = iedoc.getSectionById("internalIdentifier");
        if (null == iesec) {
            log.error ("SLUBStoragePlugin.getStreamRelativePath no section in entity of type "+entitytype +" with 'internalIdentfier' in associated iedoc found, do you use plugin for others than permanent data (IE and filestream)? You should not! Also do not use it for SIPstoragePermanentGroup!");
            throw new Exception("error, no section in entity of type "+entitytype +" with 'internalIdentfier' in associated iedoc found, do you use plugin for others than permanent data (IE and filestream)? You should not! Also do not use it for SIPstoragePermanentGroup!");
        }
        String iepid = null;
        //log.debug("SLUBStoragePlugin.getStreamRelativePath iesec=" + iesec);
        List<DnxSectionRecord> records = iesec.getRecordList();
        for (DnxSectionRecord element : records) {
            if ("PID".equals(element.getKeyById("internalIdentifierType").getValue())) {
                iepid = element.getKeyById("internalIdentifierValue").getValue(); // found IEPID
                break;
            }
        }
        // raise Exception if IEPID is null
        if (null == iepid) {
            log.error ("SLUBStoragePlugin.getStreamRelativePath iesec="+iesec );
            throw new Exception("error, could not get IEPID for storedEntityMetaData:"+storedEntityMetaData +" of type " + entitytype);
        }
        //log.debug("SLUBStoragePlugin.getStreamRelativePath iepid=" + iepid + " (entitytype=" + entitytype + ")");
        // get creationDate of "objectCharacteristics"
        String datestring = iedoc.getSectionKeyValue("objectCharacteristics", "creationDate");
        Calendar date = Calendar.getInstance();
        // date ist there stored in format (example): 2014-01-15 14:28:01
        SimpleDateFormat sdf;
        sdf = new SimpleDateFormat("yyyy-MM-dd HH:mm:ss");
        sdf.setLenient(false); /* if parse errors, do not guess about */
        Date d = sdf.parse(datestring);
        date.setTime(d);
        //log.debug("SLUBStoragePlugin.getStreamRelativePath creation Date read=" + datestring + " parsed=" + date);
        // now create path in format /yyyy/MM/dd/IEPID/
        relativeDirectoryPath = relativeDirectoryPath + new SimpleDateFormat("yyyy").format(d);
        relativeDirectoryPath = relativeDirectoryPath + File.separator;
        relativeDirectoryPath = relativeDirectoryPath + new SimpleDateFormat("MM").format(d);
        relativeDirectoryPath = relativeDirectoryPath + File.separator;
        relativeDirectoryPath = relativeDirectoryPath + new SimpleDateFormat("dd").format(d);
        relativeDirectoryPath = relativeDirectoryPath + File.separator;
        relativeDirectoryPath = relativeDirectoryPath + iepid;
        relativeDirectoryPath = relativeDirectoryPath + File.separator;
        log.info("SLUBStoragePlugin.getStreamRelativePath relativeDirectoryPath=" + relativeDirectoryPath);
        return relativeDirectoryPath;
    }

    private void softLink(String srcPath, String destPath) throws IOException
    {
        log.info("SLUBStoragePlugin.softLink srcPath='" + srcPath + "' destPath='" + destPath + "'");
        Files.createSymbolicLink(Path.of(srcPath), Path.of(destPath));
    }

    private void hardLink(String srcPath, String destPath) throws IOException
    {
        log.info("SLUBStoragePlugin.hardLink srcPath='" + srcPath + "' destPath='" + destPath + "'");
        Files.createLink(Path.of(srcPath), Path.of(destPath));
    }

    private void saveDestPathsTmpFile(String folder, String key, String path)
    {
        log.info("SLUBStoragePlugin.saveDestPathsTmpFile()");
        //log.debug("SLUBStoragePlugin.saveDestPathsTmpFile folder='" + folder + "'");
        //log.debug("SLUBStoragePlugin.saveDestPathsTmpFile key='" + key + "'");
        //log.debug("SLUBStoragePlugin.saveDestPathsTmpFile path='" + path + "'");
        if (folder == null) {
            return;
        }
        String tmpFilePath = getTempStorageDirectory(false) + "destPath";
        //log.debug("SLUBStoragePlugin.saveDestPathsTmpFile tmpFilePath='" + tmpFilePath + "'");
        File destPathDir = new File(getTempStorageDirectory(false) + "destPath" + File.separator);
        //log.debug("SLUBStoragePlugin.saveDestPathsTmpFile destPathDir='" + destPathDir + "'");
        if (!destPathDir.exists()) {
            boolean res = destPathDir.mkdirs();
            if (!res) {
                log.error("SLUBStoragePlugin.saveDestPathsTmpFile() destPathdir='" + destPathDir + "' could not be created");
            }
        }
        StorageUtil.saveDestPathToTmpFile(folder, tmpFilePath, key, path);
    }
}
